.po -3
.nh
.TH "Object" n "automatically Fri Mar 1 10:54:57 1991" "ET++ version v2.2" "ET++ class description"
.SH NAME
Object \- the root of the ET++ class inheritance tree
.SH DESCRIPTION
\fBObject\fP is the base of virtually all ET++ classes. All classes
derived from Object are called \fBregular\fP, in contrast to the
so-called built-ins like Point, Rectangle, Token, and classes like Iterator
that have derivations, but are nonetheless not regular for special reasons.
Instances of regular classes are analogously called \fBregular
objects\fP.
.br

.br
Object defines the fundamental protocols for the following concepts: Class
Descriptors, Dynamic Type-Checking, Dynamic Creation, Object Copying,
Object Input/Output, Change Propagation, and Object Comparison.
Furthermore, there are protocols covering Error Handling, Flag Handling,
and a binding to the ET++ Programming Environment. Object also defines a
virtual destructor which is extremely important.
.br
Furthermore, every regular object is registered in the global object table
within Object's constructor, and unregistered within Object's destructor.
Use of the object table is optional (see the src/makefile), but it is
required by the Inspector.
.br

.br
\fBClass Descriptors and Dynamic Type-Checks\fP
.br

.br
Every regular class has an associated class descriptor providing
information about the class at run-time. The class descriptor itself is an
regular object, namely an instance of the class Class. The most
important information in a class descriptor are the name of the class it
belongs to, and a pointer to the class descriptor of the base class. Additionally,
the class descriptor holds an so-called prototype object. Prototype objects
are particularly important for an portable implementation of the New operation
(see below).
.br
The information contained in the class descriptor is what the term meta
info refers to. Since other operations of Object's protocols require the
presence of the class descriptor, it is absolutely essential. Examples are
DeepClone (copying), ReadFrom and PrintOn (Object I/O).
.br
Every regular object can be asked for its class descriptor by IsA. IsA and
also IsKindOf allow for dynamic type-checks. Furthermore, there is the Guard
construct that provides type-safe casts. A class descriptor can be also
obtained either by the Meta construct or from the class manager, a single
global object.
.br
In order to support the ET++ Programming Environment, the class descriptor also
contains references into the source code and informations about name and
static type of the instance variables, provided they have been listed
in the meta implementation.
.br
The class descriptor for a class is set up by MetaDef and MetaImpl.
MetaDef is part of the class's declaration, and MetaImpl of the 
class's implementation. The following example shows how it is typically done:

.na
.nf
// file: A.h

#include "Object.h"

class A : public Object {
    Object *anObject;
    int aNumber;
public:
    MetaDef(A);
    A();
    ...
};

// file: A.C

...
MetaImpl(A, (TP(anObject), T(aNumber), 0));   
...

.ad b
.fi

.br

.br
\fBDynamic Creation and Object Copying\fP
.br

.br
Beside its information keeping task, the class descriptor is responsible
for the dynamic creation of new working objects. Dynamic creation is needed
since the C++ new operator requires that the class name is supplied, thus,
it is not possible to create an object without statically specifying its class
name. The New operation allows to directly ask an existing object for a new
working object of the same class.
.br
\fBObject\fP also generically implements two methods for copying objects. \fImethod Clone\fP returns a \fBshallow copy\fP, \fImethod
DeepClone\fP creates a \fBdeep copy\fP of the receiver.
.br
A shallow copy is an object that is initialized exactly the same way as the
copy constructor of its class would. Clone is part of Object's interface for
the same reason as New. The clone generally shares its part objects with the source
object (the receiver). Which referenced objects are parts, and which objects
are to be duplicated although they are parts of their containing object, depends
on the precise semantics of the class.
.br
Conceptually, the \fImethod DeepClone\fP returns a deep copy of the
receiver by recursively applying a DeepClone to all part objects of the
receiver. The generic implementation of DeepClone uses \fBObject
I/O\fP (with streams operating on a buffer in memory).
.br
The FreeAll operation is somewhat related to Object Copying. It allows to
recursively delete the part objects. As a rule-of-thumb, FreeAll is
applied to all part objects (and auxilliary objects, off course), but not
to acquainted objects. 

.br
\fBObject Input/Output\fP
.br

.br
The \fBObject I/O facility\fP writes and reads arbitrarily complex
object graphs in a textual, machine-independent format to a stream (by
operator<< and @Object::PrintOn), or from a stream (by operator>>and
@Object::ReadFrom), respectively.  ET++ Object I/O takes an half-automated
object I/O approach.
.br

The stream may operate on a file of the operating system, on a buffer in
memory, or whatever might be hidden by a stream. Object I/O properly
handles circular references. In other words, an object graph is linearized
when output, and correctly rebuilt when input.
.br
Identification of objects due to linearization during input and output is
done by operator<<, and operator>>, respectively. The identity of
objects bases on pointer identity, thus a pointer book-keeping is
maintained during object I/O operations. When during output an object is
encountered that has been already output, then an index identifying that
object is output, no more the object itself. If such an index has been read
during input, the pointer variable is assigned the pointer of the already
existing object. This is done by operator>>. If the object to be input
has not yet been constructed, operator>> allocates an object of the
appropriate class, and the object is initialized by ReadFrom. In the
current version, the pointer book-keeping is maintained by the class
manager. In future versions, this is accomplished by the streams
themselves.
.br
The primary purpose of PrintOn and ReadFrom is to apply operator<<, or
operator>> respectively, to their instance variables. Both methods are
hooks for their corresponding shift operators. Furthermore, ReadFrom has
constructor-like qualities, because the instance variables are initialized
such that a working object results. Notice that the operator>> allocates
memory, so it has a similar role like the operator new.
.br
\fBThe client has to  properly override PrintOn and ReadFrom only to get
Object I/O working\fP (assuming that the class descriptor is present).
If the client introduces new built-ins, he has to provide appropriate
overloadings of operator<< and operator>>. See the built-in class Point,
for instance.
.br

.br
Object I/O additionally supports \fBDynamic Linking\fP of classes;
when an class is encountered during input whose code is currently not
present in the image of the application's process, ET++ tries to find the
object code for that class and links it to the running application (see
\fItechnote 'Dynamic Linking'\fP; there is no further coverage in
Object's reference documentation on this topic).
.br

.br
\fBObject Comparison\fP
.br

.br

.br
If objects of a certain class are to be put into collections, then the class
should override the \fBcomparison methods\fP. Collection classes call
the \fImethod IsEqual\fP for searching and the \fImethod Compare\fP
for sorting. The \fIclass Set\fP uses the \fImethod Hash\fP for
fast access to the objects.
.br
Examples for these methods can be found in the \fIclass ByteArray\fP,
the \fIclass ObjFloat\fP or the \fIclass VObject\fP.
.br

.br
\fBChange Propagation\fP
.br

.br
The Change Propagation mechanism supports the \fBdistribution (or 
"propagation") of conceptual events\fP with a minimal coupling between the object
where the state transition happened, and the objects that shall react on this
event. Change Propagation is most often used to establish a binding
between models and views in the sense of MVC.
.br
Objects interested in changes of another object register themselves as
\fBobservers\fP of that object with the \fImethod AddObserver\fP.
An observer is unregistered by \fImethod RemoveObserver\fP. The
observers of an object can be viewed as an identity set. The observed
object \fBannounces changes\fP in its state through the \fImethod Send\fP, and the observers are notified by invocation of \fImethod DoObserve\fP. The death of the observed object is also such a
conceptual event, and observers are notified in this special case,
too.
.br
Change Messages are not explicitly modelled as objects. A change message
consists of two codes describing first the role of the changed object (so-called
id's), and second the changed aspect (so-called part codes). A value whose
type and semantics depend on the id and the part code may further describe the
change. The sender of the message is available to the observers so
they can inquire of the observed object.
.br
  
Change Propagation includes the concept of \fBDelayed Changes\fP. For
every object, there is an so-called delayed changes message counter. For
each DelayChanges call, the counter is incremented by one. The change
messages are eventually forwarded to the observers, if the delayed changes
counter is zero or gets zero after the according number of calls to
FlushChanges.  The counter is initially zero, and cannot drop below zero.
If an observed object is deleted and there are still pending change
messages, they are simply discarded. It is implementation-dependent what
happens with delayed change messages sent at a time at which the sender has
no observers.

.br

Examples:
.br
The \fIclass Text\fP uses change propagation, so an observing \fIclass TextView\fP can update the display when the text changes. The text
classes also use the delayed change mechanism.
.br
The Inspector of the ET++ Programming Environment also uses change
propagation by registering itself as an observer of the currently inspected
object.
.br

.br
\fBFlag Handling\fP
.br

.br
\fBObject\fP manages a \fBset of flags\fP which are stored
efficiently in the \fIinstvar flags\fP. Some of the flags are used
internally. New flags may be defined in subclasses.
.br
Flags for a class are defined by an enumeration type and by
means of the \fImacro BIT\fP as in the following example:
.na
.nf

enum CommandFlags {
    eCmdCanUndo        = BIT(eObjLast+1),
    eCmdCausesChange   = BIT(eObjLast+2),
    ...
    eCmdNoReplFeedback = BIT(eObjLast+8),
    eCmdLast           = eObjLast + 8
};


.ad b
.fi
The enumeration values must always be defined relative to the last flag of a
super class (in the above example \fBObject\fP). Thus, the value of the first
flag is determined this way:
.na
.nf

	- find the nearest super class which added some flags.
	- take the last enumeration value of that class.
	- increment it by one.

.ad b
.fi
In the above example the first flag for subclasses of the \fIclass Command\fP would be \fBBIT(eCmdLast+1)\fP.
.br
\fBObject\fP provides several methods to test and manipulate flags (\fImethod TestFlag\fP, \fImethod SetFlag\fP, \fImethod ResetFlag\fP
and \fImethod InvertFlag\fP). Several flags can be processed
simultaneously by "or'ing" together the enumeration values, f.i. \fBSetFlag(eCmdCanUndo|eCmdCausesChange);\fP.
.br

.br
\fBError Handling\fP
.br

.br
\fBObject\fP provides a general error handling facility which uniformly
handles program and system errors.
.br
There are 4 error handling methods (\fImethod Warning\fP, \fImethod Error\fP, \fImethod SysError\fP and \fImethod Fatal\fP) which
represent 4 different error levels indicating the severeness of the error.
These methods take the following arguments: the location (the name of the
method where the error occurred) and a format string in the style of \fBprintf\fP with optional arguments.
.br
The default error handler of ET++ prints the error message on the standard
error output. See \fItechnote 'Error Handling'\fP.
.br

.br
\fBConnection to the ET++ Programming Environment (ET++PE)\fP
.br

.br
The ET++PE depends on the \fBmetaclass information\fP (see
"Metaclasses" above) to be able to display static and dynamic
relationships between classes and objects and to find the location of the
source files of a class. Besides that, the Inspector can display instance
variables only if they have been included in the \fImacro
MetaImpl\fP.
.br
The ET++PE can additionally be supported by overriding the \fImethod
Parts\fP and the \fImethod InspectorId\fP. The \fImethod Parts\fP
allows to add important referenced objects to a list which is used by the
Object Structure Browser to show dynamic relationships between objects. The
\fImethod InspectorId\fP should return an identification string which
is displayed by the Inspector and other views of the ET++PE. For example,
the \fImethod Document::InspectorId\fP returns the document name.
.br
The Inspector and the Source Code Browser of the ET++PE can also be started
directly through \fImethod Inspect\fP and \fImethod EditSource\fP.

.br
classes are always derived from \fBObject\fP.
.br
class \fBObject\fP
is never reused directly.
.br
class \fBObject\fP is abstract.
.br
class \fBObject\fP contains 54 methods.
.PD 0
.TP 20
\fBowner of class:\fR
nobody.
.PD 0
.TP 20
\fBbaseclasses:\fR
.na
Root
.ad b
.PD 0
.TP 20
\fBsubclasses:\fR
.na
AbstractViewPolicy, Assoc, BagItem, BitSet, ByteArray, ChangeMessage, Class, ClassManager, Collection, Command, DeletedObject, EvtHandler, Ink, Mark, ObjFloat, ObjInt, ObjLink, PathLookup, Picture, ProgEnv, RegularExp, Style, SysEvtHandler, Text, TextFormatter
.ad b
.PD 0
.TP 20
\fBfriend classes:\fR
.na
Class
.ad b
.PD 0
.TP 20
\fBfriend functions:\fR
.na
_Type, operator<<, operator>>
.ad b
.PD 0
.TP 20
\fBflags:\fR
ObjFlags

.SH CLASS VARIABLES
.TP
.BR isa " (public Class *)"
refers to the class descriptor of the class.
.br
For all regular classes, \fBisa\fP is automatically created and
initialized by the \fImacro MetaDef\fP and the \fImacro
MetaImpl\fP.
.br
See \fItechnote Metaclasses\fP.
Use the \fImethod IsA\fP to access the metaclass of an object.
.br
.B Known Problems:
isa \fBmust not be accessed\fP despite its public access. It should
be protected.

.SH INSTANCE VARIABLES
.TP
.BR flags " (private u_int)"
stores the flags of an object as a bitset.
.br
The maximum number of available flags depends on the size of an \fBu_int\fP. On most machines \fBu_int\fP occupies 32 bits. In this case up to
22 user-defined flags may be defined in subclasses because 10 flags are
always reserved for internal use.
Flags can be tested with \fImethod TestFlag\fP and modified with \fImethod SetFlag\fP, \fImethod ResetFlag\fP or \fImethod
InvertFlag\fP.


.SH INSTANCE METHOD LIST

.PD 0
.TP 20
change propagation
AddObserver
.br
Changed
.br
CleanupObservers
.br
DelayChanges
.br
DestroyObserverColl
.br
DoObserve
.br
FlushChanges
.br
GetObserverIter
.br
GetObservers
.br
MakeObserverColl
.br
RemoveObserver
.br
Send
.br
SetObserverColl

.PD 0
.TP 20
client
DisplayOn

.PD 0
.TP 20
comparison
Compare
.br
Hash
.br
IsEqual

.PD 0
.TP 20
constructor
Object

.PD 0
.TP 20
conversion
AsString

.PD 0
.TP 20
copying
Clone
.br
DeepClone

.PD 0
.TP 20
creation
InitNew
.br
New

.PD 0
.TP 20
destruction
FreeAll

.PD 0
.TP 20
destructor
~Object

.PD 0
.TP 20
error handling
DoError
.br
Error
.br
Fatal
.br
SysError
.br
Warning

.PD 0
.TP 20
fire walls
AbstractMethod
.br
MayNotUse

.PD 0
.TP 20
flag handling
ClearVisited
.br
InvertFlag
.br
IsDeleted
.br
IsObserved
.br
MarkAsDeleted
.br
ResetFlag
.br
SetFlag
.br
SetVisited
.br
TestFlag

.PD 0
.TP 20
meta info
ClassName

.PD 0
.TP 20
metaclass
GetAddrOf
.br
IsA
.br
Members
.br
guard

.PD 0
.TP 20
object i/o
PrintOn
.br
PrintOnWhenObserved
.br
ReadFrom

.PD 0
.TP 20
progenv support
EditSource
.br
Inspect
.br
InspectorId
.br
Parts

.SH CATEGORIES

.na
Foundation
.ad b

.SH FILES
.PD 0
.TP 20
declaration:
\fBObject.h

.SH HISTORY
joe           - May 28 1991 - created
.br
chris         - May 29 1991 - reviewed
.br
gil           - May 31 1991 - reviewed
.br
joe, chris    - Jun  3      - minor changes
.br
T. Kofler/SBG - Sep 3 1991. - Totally reworked.
.br
.so mann/Object::AbstractMethod.n
.so mann/Object::AddObserver.n
.so mann/Object::AsString.n
.so mann/Object::Changed.n
.so mann/Object::ClassName.n
.so mann/Object::CleanupObservers.n
.so mann/Object::ClearVisited.n
.so mann/Object::Clone.n
.so mann/Object::Compare.n
.so mann/Object::DeepClone.n
.so mann/Object::DelayChanges.n
.so mann/Object::DestroyObserverColl.n
.so mann/Object::DisplayOn.n
.so mann/Object::DoError.n
.so mann/Object::DoObserve.n
.so mann/Object::EditSource.n
.so mann/Object::Error.n
.so mann/Object::Fatal.n
.so mann/Object::FlushChanges.n
.so mann/Object::FreeAll.n
.so mann/Object::GetAddrOf.n
.so mann/Object::GetObserverIter.n
.so mann/Object::GetObservers.n
.so mann/Object::Hash.n
.so mann/Object::InitNew.n
.so mann/Object::Inspect.n
.so mann/Object::InspectorId.n
.so mann/Object::InvertFlag.n
.so mann/Object::IsA.n
.so mann/Object::IsDeleted.n
.so mann/Object::IsEqual.n
.so mann/Object::IsObserved.n
.so mann/Object::MakeObserverColl.n
.so mann/Object::MarkAsDeleted.n
.so mann/Object::MayNotUse.n
.so mann/Object::Members.n
.so mann/Object::New.n
.so mann/Object::Object.n
.so mann/Object::Parts.n
.so mann/Object::PrintOn.n
.so mann/Object::PrintOnWhenObserved.n
.so mann/Object::ReadFrom.n
.so mann/Object::RemoveObserver.n
.so mann/Object::ResetFlag.n
.so mann/Object::Send.n
.so mann/Object::SetFlag.n
.so mann/Object::SetObserverColl.n
.so mann/Object::SetVisited.n
.so mann/Object::SysError.n
.so mann/Object::TestFlag.n
.so mann/Object::Warning.n
.so mann/Object::guard.n
.so mann/Object::~Object.n
